# Copyright (c) 2023 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

# Portions of this source file are derived from MonkeyType, 
# a Python library for generating static type annotations from runtime types.
# Refer to the original license and copyright included with this library.


import sys
import pathlib
import numbers
import itertools
from dataclasses import dataclass
from contextlib import contextmanager
from types import FrameType
from typing import Any, Mapping, Iterator, Sequence, Optional

import numpy as np

from monkeytype.tracing import (
    CallTraceLogger,
    get_func,
    CodeFilter,
    CallTrace,
    CallTracer,
    RETURN_VALUE_OPCODE,
    YIELD_VALUE_OPCODE,
    SUPPORTED_EVENTS,
    logger,
    EVENT_CALL,
    EVENT_RETURN,
)
from monkeytype.typing import get_type

from modelx_cython.consts import (
    MX_ASSIGN_REFS,
    FORMULA_PREF,
    is_user_defined,
    MX_SELF,
    MX_MODEL_MOD,
    MX_SPACE_MOD,
    CY_MOD,
    CY_INT_T,
)


@dataclass
class TypeInfo:
    value_type: type
    is_array: bool = False
    ndim: int = 0


class DynamicTypeInfo:
    name: str
    module: str
    class_: str
    arg_types: Mapping[str, type]  # without self
    ret_type: TypeInfo

    def __init__(self, traces: Sequence[CallTrace]) -> None:
        self.name = traces[0].func.__name__
        self.module = traces[0].func.__module__
        self.class_ = ".".join(traces[0].func.__qualname__.split(".")[:-1])
        self.arg_types = self._init_arg_types(traces)
        self.ret_type = self._init_ret_type(traces)

    def get_argtype_expr(self, arg: str) -> str:
        if arg in self.arg_types:
            return get_type_expr(self.arg_types[arg])
        else:
            return ""

    def get_rettype_expr(self) -> str:
        val_t = get_type_expr(self.ret_type.value_type)
        if self.ret_type.ndim:
            return val_t + "[" + ", ".join(":" * self.ret_type.ndim) + "]"
        else:
            return val_t

    def is_arrayable(self, sizes: Mapping[str, int]) -> bool:
        return (
            self.has_args()
            and all(t is int for t in self.arg_types.values())
            and set(self.arg_types) == set(sizes)
        )

    def get_decltype_expr(self, sizes: Mapping[str, int], rettype_expr="") -> str:
        if not rettype_expr:
            rettype_expr = self.get_rettype_expr()
        return (
            rettype_expr
            + "["
            + ", ".join([str(sizes[arg]) for arg in self.arg_types.keys()])
            + "]"
        )

    def has_args(self):
        return bool(len(self.arg_types))

    def _init_arg_types(self, traces) -> Mapping[str, type]:
        types: Mapping[str, Sequence[type]] = {}
        for trace in traces:
            for arg, typ in itertools.islice(
                trace.arg_types.items(), 1, None
            ):  # remove self
                typs = types.setdefault(arg, [])
                if typ not in typs:
                    typs.append(typ)

        return {arg: self._get_arg_type(typs) for arg, typs in types.items()}

    def _get_arg_type(self, types: Sequence[type]):
        if len(types) == 1:
            return types[0]
        else:
            lt = types[0]
            for rt in types[1:]:
                lt = comp_and_get_type(lt, rt)
            return lt

    def _init_ret_type(self, traces):
        ret_types = []
        for tr in traces:
            ret_type = self.get_value_type(tr.return_value)
            if ret_type not in ret_types:
                ret_types.append(ret_type)

        if len(ret_types) == 1:
            return ret_types[0]
        else:
            lt = ret_types[0]
            for rt in ret_types[1:]:
                lt = comp_and_get_type(lt, rt)
            return lt

    @staticmethod
    def get_value_type(value):
        if isinstance(value, np.ndarray):
            return TypeInfo(
                normalize_type(value.dtype.type), is_array=True, ndim=value.ndim
            )
        else:
            return TypeInfo(normalize_type(type(value)))


@dataclass
class RefTypeInfo:
    type_expr: str

    def __init__(self, value):
        self.type_expr = get_type_expr(type(value))


def get_type_expr(typ):
    if typ is int:
        return CY_INT_T
    elif typ is float:
        return "float"
    elif type is str:
        return "str"
    else:
        return "object"


def normalize_type(typ: type):
    if issubclass(typ, numbers.Number):
        if issubclass(typ, numbers.Rational):
            return int
        else:
            return float
    else:
        if issubclass(typ, str):
            return str
        else:
            return object


def coerce_type(lt: type, rt: type):
    if issubclass(lt, numbers.Number) and issubclass(rt, numbers.Number):
        if issubclass(lt, numbers.Rational) and issubclass(rt, numbers.Rational):
            return int
        else:
            return float
    elif issubclass(lt, str) and issubclass(rt, str):
        return str
    else:
        object


def comp_and_get_type(lt: TypeInfo, rt: TypeInfo) -> TypeInfo:
    assert lt != rt  # must be different types

    if lt.is_array and rt.is_array:
        if lt.ndim == rt.ndim:
            return TypeInfo(
                coerce_type(lt.value_type, rt.value_type), is_array=True, ndim=lt.ndim
            )
        else:
            TypeInfo(object)

    elif not lt.is_array and not lt.is_array:  # both are non-array
        return TypeInfo(coerce_type(lt.value_type, rt.value_type))

    else:  # lt or rt is array
        return TypeInfo(object)


class MxCallTracer(CallTracer):
    """Add return_value to CallTrace"""

    def __init__(
        self,
        module: str,
        logger: CallTraceLogger,
        max_typed_dict_size: int,
        code_filter: Optional[CodeFilter] = None,
        sample_rate: Optional[int] = None,
    ) -> None:
        super().__init__(logger, max_typed_dict_size, code_filter, sample_rate)
        self.module = module

    def handle_return(self, frame: FrameType, arg: Any) -> None:
        # In the case of a 'return' event, arg contains the return value, or
        # None, if the block returned because of an unhandled exception. We
        # need to distinguish the exceptional case (not a valid return type)
        # from a function returning (or yielding) None. In the latter case, the
        # the last instruction that was executed should always be a return or a
        # yield.
        typ = get_type(arg, max_typed_dict_size=self.max_typed_dict_size)
        last_opcode = frame.f_code.co_code[frame.f_lasti]
        trace = self.traces.get(frame)
        if trace is None:
            return
        elif last_opcode == YIELD_VALUE_OPCODE:
            trace.add_yield_type(typ)
            trace.return_value = arg  # Monkeytyped for modelx
        else:
            if last_opcode == RETURN_VALUE_OPCODE:
                trace.return_type = typ
                trace.return_value = arg  # Monkeytyped for modelx
            del self.traces[frame]
            self.logger.log(trace)
            if trace.func.__name__ == MX_ASSIGN_REFS:
                self.logger.members[trace.funcname] = {
                    k: v
                    for k, v in frame.f_locals[MX_SELF].__dict__.items()
                    if is_user_defined(k)
                }

    def __call__(self, frame: FrameType, event: str, arg: Any) -> "CallTracer":
        code = frame.f_code
        if (
            event not in SUPPORTED_EVENTS
            or code.co_name == "trace_types"
            or self.should_trace
            and not self.should_trace(code)
        ):
            return self

        # Filter by module name here
        func = get_func(frame)
        if not func or not func.__module__.split(".")[0] == self.module:
            return self

        try:
            if event == EVENT_CALL:
                self.handle_call(frame)
            elif event == EVENT_RETURN:
                self.handle_return(frame, arg)
            else:
                logger.error("Cannot handle event %s", event)
        except Exception:
            logger.exception("Failed collecting trace")
        return self


class MxCodeFilter:

    def __call__(self, code):
        # Since called many times, check module name and function name only
        if (
            code.co_filename[-len(MX_MODEL_MOD) - 3 : -3] == MX_MODEL_MOD
            or code.co_filename[-len(MX_SPACE_MOD) - 3 : -3] == MX_SPACE_MOD
        ):
            if code.co_name[:3] == FORMULA_PREF:
                return True
            elif code.co_name == MX_ASSIGN_REFS:
                return True
            elif code.co_name == "__call__":
                return True

        return False


class MxCallTraceLogger(CallTraceLogger):
    """Log and store/print records collected by a CallTracer."""

    def __init__(self, new_model_name: str = None) -> None:
        super().__init__()
        self.new_name = new_model_name
        self._traces = {}  # funcname -> [trace]
        self.members = {}  # (module, qualname) -> value
        self.type_info = {}  # funcname -> MethodTypeInfo
        self.ref_type_info = {}
        self.modules = []

    def log(self, trace: CallTrace) -> None:
        """Log a single call trace."""
        self._traces.setdefault(trace.funcname, []).append(trace)

    def flush(self) -> None:
        for k, v in self._traces.items():
            self.type_info[k] = info = DynamicTypeInfo(v)
            if info.module not in self.modules:
                self.modules.append(info.module)
        self._traces.clear()

        for k, v in self.members.items():
            for name, value in v.items():
                names = k.split(".")
                assert names[-1] == MX_ASSIGN_REFS
                self.ref_type_info[".".join(names[:-1] + [name])] = RefTypeInfo(value)
        self.members.clear()

        if self.new_name:
            self._change_model_name()

    def _change_model_name(self):
        """Change the model name stored in members"""

        def replace_first_name(dotted_name: str, name: str):
            names = dotted_name.split(".")
            names[0] = name
            return ".".join(names)           
        
        for i, v in enumerate(self.modules):
            self.modules[i] = replace_first_name(v, self.new_name)

        for key in list(self.type_info.keys()):
            v = self.type_info.pop(key)
            v.module = replace_first_name(v.module, self.new_name)
            self.type_info[replace_first_name(key, self.new_name)] = v

        for key in list(self.ref_type_info.keys()):
            self.ref_type_info[replace_first_name(key, self.new_name)] = self.ref_type_info.pop(key)



@contextmanager
def trace_calls(
    module: str,
    logger: CallTraceLogger,
    max_typed_dict_size: int,
    code_filter: Optional[CodeFilter] = None,
    sample_rate: Optional[int] = None,
) -> Iterator[None]:
    """Enable call tracing for a block of code"""
    old_trace = sys.getprofile()
    sys.setprofile(
        MxCallTracer(module, logger, max_typed_dict_size, code_filter, sample_rate)
    )
    try:
        yield
    finally:
        sys.setprofile(old_trace)
        logger.flush()



# Copyright (c) 2023 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union, Sequence, Mapping
try:
    from types import NoneType
except ImportError: # Python -3.9
    NoneType = type(None)
from dataclasses import dataclass
import libcst as cst
from libcst._flatten_sentinel import FlattenSentinel
from libcst._nodes.module import Module
from libcst._nodes.statement import BaseStatement, ClassDef
from libcst._removal_sentinel import RemovalSentinel
import libcst.matchers as m
from libcst.metadata import ParentNodeProvider, ScopeProvider, GlobalScope, ClassScope

from modelx_cython.config import TranslationSpec
from modelx_cython.tracer import TypeInfo, DynamicTypeInfo

from modelx_cython.consts import (
    FORMULA_PREF,
    CY_MOD,
    GLOBAL_PREF,
    VAR_PREF,
    HAS_PREF,
    SPACE_PREF,
    MX_SELF,
    MX_SYS_MOD,
    CY_BOOL_T,
    MX_ASSIGN_REFS,
    MX_COPY_REFS,
    is_user_defined,
)


class LexicalCellsInfo:
    module_name: str
    cls_name: str
    name: str
    params: Sequence[str]

    def __init__(self, module_name, cls_name, name, params, spec) -> None:
        self.module_name: str = module_name
        self.cls_name: str = cls_name
        self.name: str = name
        self.params: Sequence[str] = params
        self.spec = spec

    @property
    def keystr(self):
        pref = "" if self.is_special() else FORMULA_PREF
        return self.module_name + "." + self.cls_name + "." + pref + self.name

    def is_special(self):
        return self.name[:2] == self.name[-2:] == "__"


class CombinedCellsInfo(LexicalCellsInfo):
    _typeinfo: DynamicTypeInfo

    def __init__(self, cellsinfo, typeinfo) -> None:
        super().__init__(
            cellsinfo.module_name, cellsinfo.cls_name, cellsinfo.name, cellsinfo.params, cellsinfo.spec
        )
        self._typeinfo = typeinfo

    def has_typeinfo(self):
        return bool(self._typeinfo)

    def has_args(self):
        return bool(self.params)

    def get_argtype_expr(self, arg: str) -> str:
        return self._typeinfo.get_argtype_expr(arg)

    def get_rettype_expr(self):

        ret_t = self.spec.get(TranslationSpec.RET_T)
        if ret_t:
            return ret_t
        elif self.has_typeinfo():
            return self._typeinfo.get_rettype_expr()
        else:
            return "object"

    def is_arrayable(self, sizes):
        if self.has_typeinfo():
            return self._typeinfo.is_arrayable(sizes)
        else:
            return False

    def get_decltype_expr(self, sizes: Mapping[str, int], rettype_expr=""):
        return self._typeinfo.get_decltype_expr(sizes, rettype_expr=rettype_expr)


@dataclass
class RefInfo:
    module_name: str
    cls_name: str
    name: str
    type_expr: Union[str, NoneType]

class SpaceAddin:

    def get_parent(self, node, level=0):
        while level:
            node = self.get_metadata(ParentNodeProvider, node)
            level -= 1
        return node
    
    def get_scope(self, node, level=0):
        return self.get_metadata(ScopeProvider, self.get_parent(node, level=level))
    
    def is_space_scope(self, node, level=0):
        scope = self.get_scope(node, level)
        return  bool(
            isinstance(scope, ClassScope)
            and scope.name[: len(SPACE_PREF)] == SPACE_PREF
            and isinstance(scope.parent, GlobalScope)
        )

class SpaceVisitor(m.MatcherDecoratableVisitor, SpaceAddin):
    METADATA_DEPENDENCIES = (ScopeProvider, ParentNodeProvider)

    def __init__(self, module_name, source, spec, type_info: dict, ref_type_info: dict):
        super().__init__()
        self.module_name = module_name
        self.spec = spec
        self.cells_info = {}
        self.ref_info = {}
        self.space_info = {}
        self.type_info = type_info
        self.ref_type_info = ref_type_info
        self.wrapper = cst.metadata.MetadataWrapper(cst.parse_module(source))
        self.wrapper.visit(self)


    @m.call_if_inside(m.ClassDef())
    @m.call_if_inside(m.FunctionDef(name=cst.Name("__init__")))
    @m.leave(m.SimpleStatementLine())
    def collect_space_info(self, original_node):
        if self.is_space_scope(original_node, level=2):
            node = original_node

            try:
                target = cst.ensure_type(node.body[0], cst.Assign).targets[0].target
            except Exception:
                return
            
            if (target.value.value == MX_SELF and is_user_defined(target.attr.value)):
                self.space_info[target.attr.value] = self.wrapper.module.code_for_node(node.body[0].value)

    @m.call_if_inside(m.ClassDef())
    @m.call_if_inside(m.FunctionDef(name=cst.Name(MX_ASSIGN_REFS)))
    @m.leave(m.SimpleStatementLine())
    def collect_refs_info(self, original_node):

        if self.is_space_scope(original_node, level=2):
            # SimpleStatemetn in IndentedBlock in FunctionDef in IndentedBlock in ClassDef

            node = original_node
            for _ in range(4):
                node = self.get_metadata(ParentNodeProvider, node)

            cls_name = cst.ensure_type(node, cst.ClassDef).name.value

            try:
                name = cst.ensure_type(
                    cst.ensure_type(original_node.body[0], cst.Assign).targets[0],
                    cst.AssignTarget,
                ).target.attr.value
            except Exception:   # igonore other than assignments, such as 'pass'
                return

            ref_type_info = self.ref_type_info.get(
                self.module_name + "." + cls_name + "." + name, None
            )
            self.ref_info[self.module_name, cls_name, name] = RefInfo(
                self.module_name,
                cls_name,
                name,
                ref_type_info.type_expr if ref_type_info else None,
            )

    @m.call_if_inside(m.ClassDef())
    @m.visit(m.FunctionDef())
    def collect_methods(self, original_node):

        if self.is_space_scope(original_node):
            cls_name = cst.ensure_type(
                self.get_parent(original_node, level=2),
                cst.ClassDef,
            ).name.value

            if original_node.name.value[: len(FORMULA_PREF)] == FORMULA_PREF:
                # _f_ methods
                pass
            elif original_node.name.value[: len(GLOBAL_PREF)] == GLOBAL_PREF:
                # _mx_ methods
                pass
            elif (
                original_node.name.value[:2] == "__"
                and original_node.name.value != "__call__"
            ):
                # Special methods
                pass
            else:
                # cells
                name = original_node.name.value
                params = [
                    p.name.value
                    for p in original_node.params.params
                    + original_node.params.posonly_params
                    if p.name.value != MX_SELF
                ]
                spec = self.spec.get_spec(self.module_name + "." + cls_name).get("cells", {}).get(name, {})

                ci = LexicalCellsInfo(
                    module_name=self.module_name,
                    cls_name=cls_name,
                    name=name,
                    params=params,
                    spec=spec
                )
                self.cells_info[cls_name, name] = CombinedCellsInfo(
                    ci, self.type_info.get(ci.keystr, None)
                )

        return False


class SpaceTransformer(m.MatcherDecoratableTransformer, SpaceAddin):
    METADATA_DEPENDENCIES = (ScopeProvider, ParentNodeProvider)

    def __init__(
        self,
        module_name: str,
        source: str,
        type_info: dict,
        ref_type_info: dict,
        spec: TranslationSpec,
    ) -> None:
        super().__init__()
        self.module_name = module_name
        self.wrapper = cst.metadata.MetadataWrapper(cst.parse_module(source))
        self.module = self.wrapper.module
        self.spec = spec
        space = SpaceVisitor(module_name, source, spec, type_info, ref_type_info)
        self.cells_info = space.cells_info
        self.ref_info = space.ref_info
        self.space_info = space.space_info

    @property
    def package(self) -> str:
        return self.module_name.split(".")[0]

    @property
    def transformed(self):
        return self.wrapper.visit(self)

    def get_arg_sizes(self, cls_name: str) -> Mapping[str, int]:
        space = self.spec.get_spec(self.module_name + "." + cls_name)
        params = space.get("cells_params", {})
        return {k: v["size"] for k, v in params.items() if "size" in v}


    def leave_Module(self, original_node: Module, updated_node: Module) -> Module:
        return updated_node.with_changes(
            body=(
                cst.parse_statement(
                    f"from cython.cimports.{self.package} import {MX_SYS_MOD}",
                    config=updated_node.config_for_parsing,
                ),
                cst.parse_statement(
                    f"import cython as {CY_MOD}", config=updated_node.config_for_parsing
                ),
                *updated_node.body,
            )
        )

    def leave_ClassDef(
        self, original_node: ClassDef, updated_node: ClassDef
    ) -> Union[BaseStatement, FlattenSentinel[BaseStatement], RemovalSentinel]:
        cls_name: str = original_node.name.value
        if cls_name[: len(SPACE_PREF)] == SPACE_PREF and isinstance(
            self.get_metadata(ScopeProvider, original_node), GlobalScope
        ):
            decl_stmts = []
            for cells in self.cells_info.values():
                if cells.is_special():
                    pass
                elif cells.has_args():
                    if cells.is_arrayable(self.get_arg_sizes(cls_name)):
                        decl_stmts.append(
                            cst.parse_statement(
                                VAR_PREF
                                + cells.name
                                + ": "
                                + cells.get_decltype_expr(self.get_arg_sizes(cls_name)),
                                config=self.module.config_for_parsing,
                            )
                        )
                        decl_stmts.append(
                            cst.parse_statement(
                                HAS_PREF
                                + cells.name
                                + ": "
                                + cells.get_decltype_expr(
                                    self.get_arg_sizes(cls_name),
                                    rettype_expr=CY_BOOL_T,
                                ),
                                config=self.module.config_for_parsing,
                            )
                        )
                    else:
                        decl_stmts.append(
                            cst.parse_statement(
                                VAR_PREF + cells.name + ": dict",
                                config=self.module.config_for_parsing,
                            )
                        )
                else:
                    rettype = cells.get_rettype_expr()
                    decl_stmts.append(
                        cst.parse_statement(
                            VAR_PREF + cells.name + ": " + rettype,
                            config=self.module.config_for_parsing,
                        )
                    )

                    decl_stmts.append(
                        cst.parse_statement(
                            HAS_PREF + cells.name + ": " + CY_BOOL_T,
                            config=self.module.config_for_parsing,
                        )
                    )

            is_first = True
            for ref in self.ref_info.values():
                stmt = cst.parse_statement(
                    f"{ref.name} = {CY_MOD}.declare({ref.type_expr}, visibility='public')",
                    config=self.module.config_for_parsing,
                )
                if is_first:
                    stmt = stmt.with_changes(
                        leading_lines=tuple(stmt.leading_lines) + (cst.EmptyLine(),)
                    )
                    is_first = False

                decl_stmts.append(stmt)

            is_first = True
            for space in self.space_info.keys():

                stmt = cst.parse_statement(
                    f"{space} = {CY_MOD}.declare('object', visibility='public')",
                    config=self.module.config_for_parsing,
                )
                if is_first:
                    stmt = stmt.with_changes(
                        leading_lines=tuple(stmt.leading_lines) + (cst.EmptyLine(),)
                    )
                    is_first = False

                decl_stmts.append(stmt)


            decorator = cst.Decorator(
                decorator=cst.Attribute(value=cst.Name(CY_MOD), attr=cst.Name("cclass"))
            )

            if decl_stmts:
                # Add blank lines below classdef
                decl_stmts[0] = decl_stmts[0].with_changes(
                    leading_lines=tuple(decl_stmts[0].leading_lines) + (cst.EmptyLine(),)
                )
                indented_block = cst.ensure_type(
                    updated_node.body, cst.IndentedBlock
                ).with_changes(body=tuple(decl_stmts) + updated_node.body.body)
                return updated_node.with_changes(
                    decorators=(decorator,), body=indented_block
                )
            else:
                return updated_node.with_changes(decorators=(decorator,))
        else:
            return updated_node

    @m.call_if_inside(m.ClassDef())
    @m.call_if_inside(m.FunctionDef(name=cst.Name("__init__")))
    @m.leave(m.SimpleStatementLine())
    def remove_cache_assigns(self, original_node, updated_node):
        funcdef = self.get_parent(original_node, level=2)
        clsdef = self.get_parent(funcdef, level=2)
        if (
            m.matches(funcdef, m.FunctionDef(name=cst.Name("__init__")))
            and m.matches(clsdef, m.ClassDef())
            and clsdef.name.value[: len(SPACE_PREF)] == SPACE_PREF
            and m.matches(original_node, m.SimpleStatementLine(body=(m.Assign(),)))
            and m.matches(
                original_node.body[0],
                m.Assign(
                    targets=(m.AssignTarget(target=m.Attribute(value=m.Name(MX_SELF))),)
                ),
            )
            and (
                original_node.body[0].targets[0].target.attr.value[: len(HAS_PREF)]
                == HAS_PREF
                or original_node.body[0].targets[0].target.attr.value[: len(VAR_PREF)]
                == VAR_PREF
            )
        ):
            return cst.RemoveFromParent()

        return updated_node

    def _add_param_type_hints(
        self, funcdef: cst.FunctionDef, cls_name: str
    ) -> Union[cst.Parameters, NoneType]:
        param_list = list(funcdef.params.params + funcdef.params.posonly_params)[
            1:
        ]  # remove self

        name = funcdef.name.value
        if name[:len(FORMULA_PREF)] == FORMULA_PREF:
            name = name[len(FORMULA_PREF):]

        cells = self.cells_info.get((cls_name, name))

        # Add parameter type hints
        if cells and cells.has_typeinfo() and cells.has_args():
            updated_params = [funcdef.params.params[0]]  # add self first
            for param in param_list:
                param_name = param.name.value
                if cells.get_argtype_expr(param_name):
                    updated_params.append(
                        param.with_changes(
                            annotation=cst.Annotation(
                                annotation=cst.parse_expression(
                                    cells.get_argtype_expr(param_name),
                                    config=self.module.config_for_parsing,
                                )
                            )
                        )
                    )
                else:
                    updated_params.append(param)

            return funcdef.params.with_changes(params=tuple(updated_params))

        return None


    @m.call_if_inside(m.ClassDef())
    @m.call_if_inside(m.FunctionDef(name=cst.Name(MX_COPY_REFS)))
    @m.call_if_inside(m.SimpleStatementLine())
    @m.call_if_inside(m.Assign())
    @m.call_if_inside(m.Attribute())
    @m.leave(m.Name(value="base"))
    def rename_base(self, original_node, updated_node):
        return updated_node.with_changes(value="base_")

    @m.call_if_inside(m.ClassDef())
    @m.leave(m.FunctionDef())
    def update_method(self, original_node, updated_node):

        if self.is_space_scope(original_node):
            cls_name = cst.ensure_type(
                self.get_parent(original_node, level=2),
                cst.ClassDef,
            ).name.value

            meth_name = original_node.name.value

            if meth_name[: len(FORMULA_PREF)] == FORMULA_PREF:
                # _f_ methods
                cells = self.cells_info.get((cls_name, meth_name[len(FORMULA_PREF):]))

                decorators = [
                    cst.Decorator(
                        decorator=cst.Attribute(
                            value=cst.Name(CY_MOD), attr=cst.Name("cfunc")
                        )
                    )
                ]
                if cells and cells.has_typeinfo():
                    returns = cst.Annotation(
                        annotation=cst.parse_expression(
                            cells.get_rettype_expr(),
                            config=self.module.config_for_parsing,
                        )
                    )
                    if cells.has_args():
                        parameters = self._add_param_type_hints(
                            updated_node, cls_name=cls_name
                        )
                        return updated_node.with_changes(
                            decorators=decorators, params=parameters, returns=returns
                        )
                    else:
                        return updated_node.with_changes(
                            decorators=decorators, returns=returns
                        )
                else:
                    return updated_node.with_changes(decorators=decorators)

            elif meth_name == MX_COPY_REFS:
                decorators = [
                    cst.Decorator(
                        decorator=cst.Attribute(
                            value=cst.Name(CY_MOD), attr=cst.Name("ccall")
                        )
                    )
                ]

                decl_stmt = cst.parse_statement(
                    f"base_: {cls_name} = {CY_MOD}.cast({cls_name}, base)",
                    config=self.module.config_for_parsing,
                )
                decl_stmt = decl_stmt.with_changes(leading_lines=(cst.EmptyLine(),))
                stmts = list(cst.ensure_type(updated_node.body, cst.IndentedBlock).body)
                stmts.insert(0, decl_stmt)
                indented_block = updated_node.body.with_changes(body=stmts)

                return updated_node.with_changes(
                    decorators=decorators, body=indented_block
                )

            elif meth_name[: len(GLOBAL_PREF)] == GLOBAL_PREF:
                # other _mx_ methods
                return updated_node

            elif meth_name == "__call__":
                # Special methods
                cells = self.cells_info.get((cls_name, meth_name))

                if cells and cells.has_typeinfo() and cells.has_args():
                    parameters = self._add_param_type_hints(
                        updated_node, cls_name=cls_name
                    )
                    return updated_node.with_changes(params=parameters)

            elif meth_name[:2] == "__":
                # Special methods
                pass

            else:
                # cells
                cells: CombinedCellsInfo = self.cells_info[cls_name, meth_name]

                decorators = [
                    cst.Decorator(
                        decorator=cst.Attribute(
                            value=cst.Name(CY_MOD), attr=cst.Name("ccall")
                        )
                    )
                ]

                if cells.has_typeinfo():
                    # Return type
                    returns = cst.Annotation(
                        annotation=cst.parse_expression(
                            cells.get_rettype_expr(),
                            config=self.module.config_for_parsing,
                        )
                    )

                    if cells.is_arrayable(self.get_arg_sizes(cls_name)):
                        # Add parameter type hints
                        parameters = self._add_param_type_hints(
                            updated_node, cls_name=cls_name
                        )
                        # Construct indented_block to replace the original one
                        if_expr = f"{MX_SELF}.{HAS_PREF}{meth_name}[{', '.join(cells.params)}]"
                        expr_node = cst.parse_expression(
                            if_expr, config=self.module.config_for_parsing
                        )
                        stmt_node = cst.parse_statement(
                            if_expr + " = True", config=self.module.config_for_parsing
                        )
                        # updated_node.body.body[0].test
                        # FunctionDef.body: IndentedBlock
                        # IndentedBlock.body: tuple(If,)
                        # If.test: Expr
                        # If.orelse: Else
                        # Else.body: IndentedBlock
                        # IndentedBlock.body: tuple(SimpleStatementLine,...)
                        if_node = cst.ensure_type(updated_node.body.body[0], cst.If)
                        stmts = list(
                            cst.ensure_type(if_node.orelse, cst.Else).body.body
                        )
                        stmts.insert(-1, stmt_node)

                        if_node = if_node.with_changes(
                            test=expr_node,
                            orelse=if_node.orelse.with_changes(
                                body=if_node.orelse.body.with_changes(body=stmts)
                            ),
                        )
                        indented_block = cst.ensure_type(
                            updated_node.body, cst.IndentedBlock
                        ).with_changes(body=(if_node,))

                        return updated_node.with_changes(
                            decorators=decorators,
                            params=parameters,
                            returns=returns,
                            body=indented_block,
                        )

                    else:
                        return updated_node.with_changes(
                            decorators=decorators, returns=returns
                        )
                else:
                    return updated_node.with_changes(decorators=decorators)

        return updated_node



# Copyright (c) 2023 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.


FILE_PREF = "_mx_"
GLOBAL_PREF = "_mx_"
MODULE_PREF = "_m_"
SPACE_PREF = "_c_"
FORMULA_PREF = "_f_"
VAR_PREF = "_v_"
HAS_PREF = "_has_"
MX_SELF = "self"

MX_MODEL_MOD = FILE_PREF + "model"
MX_SPACE_MOD = FILE_PREF + "classes"
MX_SYS_MOD = FILE_PREF + "sys"

MX_ASSIGN_REFS = GLOBAL_PREF + "assign_refs"
MX_COPY_REFS = GLOBAL_PREF + "copy_refs"
CY_MOD = GLOBAL_PREF + "cy"
NP_MOD = GLOBAL_PREF + "np"
PD_MOD = GLOBAL_PREF + "pd"

ANY_T = "object"
FLOAT_T = "float"
CY_BOOL_T = CY_MOD + ".bint"
CY_INT_T = CY_MOD + ".int"
CY_FLOAT_T = CY_MOD + ".double"


def is_user_defined(name: str):
    if name[:1] == "_":
        return False
    else:
        return True


